import chalk from 'chalk';

import enumRuleIsActive from './enum-rule-is-active';
import format from './format';
import getForcedCaseFn from './get-forced-case-fn';
import getForcedLeadingFn from './get-forced-leading-fn';
import getHasName from './get-has-name';
import meta from './meta';

/**
 * Get a cli prompt based on rule configuration
 * @param  {string} type     type of the data to gather
 * @param  {object} context     rules to parse
 * @return {object}          prompt instance
 */
// eslint-disable-next-line complexity
function getPrompt(type, context = {}) {
    const { rules = [], settings = {}, results = {}, prompter } = context;

    if (typeof prompter !== 'function') {
        throw new TypeError('Missing prompter function in getPrompt context');
    }

    const prompt = prompter();

    if (typeof prompt.removeAllListeners !== 'function') {
        throw new TypeError('getPrompt: prompt.removeAllListeners is not a function');
    }

    if (typeof prompt.command !== 'function') {
        throw new TypeError('getPrompt: prompt.command is not a function');
    }

    if (typeof prompt.catch !== 'function') {
        throw new TypeError('getPrompt: prompt.catch is not a function');
    }

    if (typeof prompt.addListener !== 'function') {
        throw new TypeError('getPrompt: prompt.addListener is not a function');
    }

    if (typeof prompt.log !== 'function') {
        throw new TypeError('getPrompt: prompt.log is not a function');
    }

    if (typeof prompt.delimiter !== 'function') {
        throw new TypeError('getPrompt: prompt.delimiter is not a function');
    }

    if (typeof prompt.show !== 'function') {
        throw new TypeError('getPrompt: prompt.show is not a function');
    }

    const enumRule = rules.filter(getHasName('enum')).filter(enumRuleIsActive)[0];

    const emptyRule = rules.filter(getHasName('empty'))[0];

    const mustBeEmpty = emptyRule
        ? emptyRule[1][0] > 0 && emptyRule[1][1] === 'always'
        : false;

    const mayNotBeEmpty = emptyRule
        ? emptyRule[1][0] > 0 && emptyRule[1][1] === 'never'
        : false;

    const mayBeEmpty = !mayNotBeEmpty;

    if (mustBeEmpty) {
        prompt.removeAllListeners('keypress');
        prompt.removeAllListeners('client_prompt_submit');
        prompt.ui.redraw.done();
        return Promise.resolve();
    }

    const caseRule = rules.filter(getHasName('case'))[0];

    const forceCaseFn = getForcedCaseFn(caseRule);

    const leadingBlankRule = rules.filter(getHasName('leading-blank'))[0];

    const forceLeadingBlankFn = getForcedLeadingFn(leadingBlankRule);

    const maxLenghtRule = rules.filter(getHasName('max-length'))[0];

    const hasMaxLength = maxLenghtRule && maxLenghtRule[1][0] > 0;

    const inputMaxLength = hasMaxLength ? maxLenghtRule[1][1] : Infinity;

    const headerLength = settings.header ? settings.header.length : Infinity;

    const remainingHeaderLength = headerLength
        ? headerLength -
        [
            results.type,
            results.scope,
            results.scope ? '()' : '',
            results.type && results.scope ? ':' : '',
            results.subject,
        ].join('').length
        : Infinity;

    const maxLength = Math.min(inputMaxLength, remainingHeaderLength);

    return new Promise(resolve => {
        // Add the defined enums as sub commands if applicable
        if (enumRule) {
            const [, [, , enums]] = enumRule;

            enums.forEach(enumerable => {
                const enumSettings = (settings.enumerables || {})[enumerable] || {};
                prompt
                    .command(enumerable)
                    .description(enumSettings.description || '')
                    .action(() => {
                        prompt.removeAllListeners();
                        prompt.ui.redraw.done();
                        return resolve(forceLeadingBlankFn(forceCaseFn(enumerable)));
                    });
            });
        } else {
            prompt.catch('[text...]').action(parameters => {
                const { text = '' } = parameters;
                prompt.removeAllListeners();
                prompt.ui.redraw.done();
                return resolve(forceLeadingBlankFn(forceCaseFn(text.join(' '))));
            });
        }

        if (mayBeEmpty) {
            // Add an easy exit command
            prompt
                .command('skip')
                .description('如果可能的话，忽略当前输入选项。')
                .action(() => {
                    prompt.removeAllListeners();
                    prompt.ui.redraw.done();
                    resolve('');
                });
        }

        // Handle empty input
        const onSubmit = input => {
            if (input.length > 0) {
                return;
            }

            // Show help if enum is defined and input may not be empty
            if (mayNotBeEmpty) {
                prompt.ui.log(chalk.yellow(`⚠ ${chalk.bold(type)} 选项不能为空，请重新输入。\n`));
            }

            if (mayBeEmpty) {
                prompt.ui.log(chalk.cyan(`ℹ 输入 ${chalk.bold('skip')} 可以忽略 ${chalk.bold(type)} 选项。\n`));
            }

            if (enumRule) {
                const [, [, , enums]] = enumRule;
                prompt.log(`  仅支持以下 ${chalk.bold(type)} 选项:\n`);
                enums.forEach(enumerable => {
                    const enumSettings = (settings.enumerables || {})[enumerable] || {};
                    prompt.log(`    ${enumerable.padEnd(8)}               ${enumSettings.description}`);
                });
                prompt.log('\n');
            }
        };

        const drawRemaining = length => {
            if (length < Infinity) {
                const colors = [
                    {
                        'threshold': 5,
                        'color': 'red',
                    },
                    {
                        'threshold': 10,
                        'color': 'yellow',
                    },
                    {
                        'threshold': Infinity,
                        'color': 'grey',
                    },
                ];

                const color = colors
                    .filter(item => item.threshold >= length)
                    .map(item => item.color)[0];

                prompt.ui.redraw(chalk[color](`剩余可输入 ${length} 个字符`));
            }
        };

        const onKey = event => {
            const sanitized = forceCaseFn(event.value);
            const cropped = sanitized.slice(0, maxLength);

            // We **could** do live editing, but there are some quirks to solve
            /* const live = merge({}, results, {
                [type]: cropped
            });
            prompt.ui.redraw(`\n\n${format(live, true)}\n\n`); */

            if (maxLength) {
                drawRemaining(maxLength - cropped.length);
            }
            prompt.ui.input(cropped);
        };

        prompt.addListener('keypress', onKey);
        prompt.addListener('client_prompt_submit', onSubmit);

        prompt.log(
            `\n\n请输入 ${chalk.bold(type)} 选项: ${chalk.green(`${settings.description}`)} ${meta({
                '可选': !mayNotBeEmpty,
                '必填': mayNotBeEmpty,
                'tab 补全': typeof enumRule !== 'undefined',
                '头部信息': typeof settings.header !== 'undefined',
                '支持多行': settings.multiline,
            })}\n`,
        );

        prompt.log(`\n${format(results, true)}\n`);

        drawRemaining(maxLength);

        prompt.delimiter(`❯ ${type}:`).show();
    });
}

export default getPrompt;
